import os
import sys
import threading

os.system("")  # enables ansi escape characters in terminal

CODE = {
    "CYAN": "\033[36m",
    "END": "\033[0m",
    "RM_LINE": "\033[2K\r",
    "HIDE_CURSOR": "\033[?25l",
    "SHOW_CURSOR": "\033[?25h",
    "CURSOR_NEXT_LINE": "\033[1E",
    "CURSOR_INIT": "\033[0G",
}


class ProgressBase(threading.Thread):
    inplace = None
    stopFlag = None

    def __init__(self):
        self.rlock = threading.RLock()
        self.cv = threading.Condition()
        threading.Thread.__init__(self)
        self.setDaemon(True)

    def __call__(self):
        self.start()

    def back_step(self):
        if self.inplace:
            sys_print(CODE["CURSOR_NEXT_LINE"])

    def remove_line(self):
        if self.inplace:
            sys_print(CODE["RM_LINE"])

    def start(self):
        self.stopFlag = 0
        threading.Thread.start(self)

    def stop(self):
        self.stopFlag = 1
        sys_print(CODE["SHOW_CURSOR"])
        self.cv.acquire()
        self.cv.notify()
        self.cv.release()
        self.rlock.acquire()


class Spinner(ProgressBase):

    def __init__(self, msg="", speed=0.1):
        self.__seq = ["⣾", "⣷", "⣯", "⣟", "⡿", "⢿", "⣻", "⣽"]
        self.__speed = speed
        self.__msg = msg
        self.inplace = 1
        ProgressBase.__init__(self)

    def run(self):
        self.rlock.acquire()
        self.cv.acquire()
        sys_print(CODE["HIDE_CURSOR"])
        while 1:
            for char in self.__seq:
                self.cv.wait(self.__speed)
                if self.stopFlag:
                    self.back_step()
                    try:
                        return
                    finally:
                        self.rlock.release()
                if self.inplace:
                    sys_print(
                        f"{CODE['CYAN']}{char}{CODE['END']} {self.__msg}{CODE['CURSOR_INIT']}"
                    )


def sys_print(msg):
    sys.stdout.write(msg)
    sys.stdout.flush()


def spin_it(function, msg, speed=0.25, *args, **kwargs):
    indicator = Spinner(msg, speed)
    indicator.start()
    result = function(*args, **kwargs)
    indicator.stop()
    indicator.remove_line()
    return result



from teradataml import DataFrame
from typing import List
from .connections import execute_sql

import json
import sys

DEFAULT_STO_MODELS_TABLE = "vmo_sto_models"


def save_evaluation_metrics(partition_df: DataFrame, metrics: List):
    """
    :param partition_df: teradata dataframe containing at least ["partition_id", "partition_metadata", "num_rows"]
    :param metrics: list of metrics to normalize and report
    :return: None
    """
    total_rows = int(partition_df.select(["num_rows"]).sum().to_pandas().iloc[0])

    metrics_sql = [
        f"SUM(CAST(partition_metadata AS JSON).JSONExtractValue('$.metrics.{metric}') *"
        f" num_rows/{total_rows}) AS {metric}"
        for metric in metrics
    ]
    joined_metrics_sql = ",".join(metrics_sql)
    metrics = DataFrame.from_query(
        f"SELECT {joined_metrics_sql} FROM {partition_df._table_name}"
    ).to_pandas()

    metrics = {metric: "{:.2f}".format(metrics.iloc[0][metric]) for metric in metrics}

    with open("artifacts/output/metrics.json", "w+") as f:
        json.dump(metrics, f, indent=2)


def save_metadata(partition_df: DataFrame):
    """
    create statistic summaries based on the provided dataframe produced via training or evaluation

    partitions.json is {
        "<partition1 key>": <partition1_metadata>,
        "<partition2 key>": <partition2_metadata>,
        ...
    }

    data_stats.json is {
        "num_rows": <num_rows>,
        "num_partitions": <num_partitions>
    }

    :param partition_df: teradata dataframe containing at least ["partition_id", "partition_metadata", "num_rows"]
    :return: None
    """

    total_rows = int(partition_df.select(["num_rows"]).sum().to_pandas().iloc[0])

    metadata_df = partition_df.select(
        ["partition_id", "partition_metadata", "num_rows"]
    ).to_pandas()

    metadata_dict = {
        r["partition_id"]: json.loads(r["partition_metadata"])
        for r in metadata_df.to_dict(orient="records")
    }

    with open("artifacts/output/partitions.json", "w+") as f:
        json.dump(metadata_dict, f, indent=2)

    data_metadata = {
        "num_rows": total_rows,
        "num_partitions": int(metadata_df.shape[0]),
    }

    with open("artifacts/output/data_stats.json", "w+") as f:
        json.dump(data_metadata, f, indent=2)


def cleanup_cli(model_version: str, models_table: str = DEFAULT_STO_MODELS_TABLE):
    """
    cli uses model version of "cli" always. We need to cleanup models table between runs.
    A better solution would be for the cli to write to a different table completely and just "recreate" on each run

    :param model_version: the model version being executed
    :param models_table: the models table for cleanup (default is vmo_sto_models)
    :return: None
    """
    if model_version == "cli":
        execute_sql(
            "DELETE FROM {table} WHERE model_version='cli'".format(table=models_table)
        )


def check_sto_version(python_interpreter="tdpython3"):
    """
    Check Python version In-Vantage against the version where this function is running,
    if it's incompatible raise an exception

    :param python_interpreter: path to python interpreter on Vantage node (default is tdpython3)
    :return: None
    """
    version_query = f"""
    SEL DISTINCT ver
    FROM SCRIPT(
        SCRIPT_COMMAND('{python_interpreter} -c "import sys; print(\\".\\".join(map(str, sys.version_info[0:2])))"')
        RETURNS('ver VARCHAR(10)') 
    );
    """
    local_version = ".".join(map(str, sys.version_info[0:2]))
    result = execute_sql(version_query)
    if result.rowcount != 1:
        raise Exception(
            "Different STO configuration on different nodes, please contact your system"
            " administrator"
        )
    try:
        remote_version = next(iter(result))[0]
    except Exception as err:
        raise Exception(
            "Different STO configuration on different nodes, please contact your"
            f" system administrator. \nError: {err}"
        )

    if local_version != remote_version:
        raise Exception(
            "Python versions not matching, local: {local}, In-Vantage: {remote}".format(
                local=local_version, remote=remote_version
            )
        )


def collect_sto_versions(
    raise_diff_config_exception=True, python_interpreter="tdpython3"
):
    """
    Collects Python and packages information from In-Vantage installation

    :param python_interpreter: path to python interpreter on Vantage node (default is tdpython3)
    :param raise_diff_config_exception: whether raise an exception if different Python versions are detected on different AMPs of Vantage system
    :return: Dict with python_version and packages versions
    """
    python_version_query = f"""
    SEL DISTINCT ver
    FROM SCRIPT(
        SCRIPT_COMMAND('{python_interpreter} -c "import sys; print(sys.version.replace(\\"\\n\\",\\" \\"))"')
        RETURNS('ver VARCHAR(100)') 
    );
    """
    result = execute_sql(python_version_query)
    if result.rowcount != 1 and raise_diff_config_exception:
        raise Exception(
            "Different STO configuration on different nodes, please contact your system"
            " administrator"
        )
    try:
        python_version = next(iter(result))[0]
    except Exception as err:
        raise Exception(
            "Different STO configuration on different nodes, please contact your"
            f" system administrator. \nError: {err}"
        )

    packages_version_query = f"""
    SEL DISTINCT pkg
    FROM SCRIPT(
        SCRIPT_COMMAND('{python_interpreter} -c "import pkg_resources; [print(pkg) for pkg in pkg_resources.working_set]"')
        RETURNS('pkg VARCHAR(100)') 
    );
    """
    result = execute_sql(packages_version_query)
    packages = {}
    for row in result:
        pair = row[0].split(" ")
        packages[pair[0]] = pair[1]

    return {"python_version": python_version, "packages": packages}


def get_joined_models_df(
    data_table: str,
    model_artefacts_table: str,
    model_version: str,
    partition_id: str = "partition_id",
):
    """
    Joins the dataset which is to be used for scoring/evaluation with the model artefacts and appends the model_artefact
    to the first row with the column name 'model'.

    Args:
        data_table: the table/view of the dataset to join
        model_artefacts_table: the model artefacts table where the model artefacts are stored
        model_version: the model version to use from the model artefacts
        partition_id: the dataset partition_id

    Returns:
        DataFrame
    """
    query = f"""
    SELECT d.*, CASE WHEN n_row=1 THEN m.model ELSE null END AS model 
        FROM (SELECT x.*, ROW_NUMBER() OVER (PARTITION BY x.{partition_id} ORDER BY x.{partition_id}) AS n_row FROM {data_table} x) AS d
        LEFT JOIN {model_artefacts_table} m
        ON d.{partition_id} = m.partition_id
        WHERE m.model_version = '{model_version}'
    """

    return DataFrame.from_query(query)


def get_joined_single_model_df(
    data_table: str,
    model_version: str,
    model_artefacts_table: str = DEFAULT_STO_MODELS_TABLE,
    partition_id: str = "partition_id",
):
    """
    Joins the dataset which is to be used for scoring/evaluation with a single model artefact and appends that artefact
    to the first row of each partition with the column name 'model'.
    If there's more than one artefact corresponding to 'model_version' then the random first one is taken.


    Args:
        data_table: the table/view of the dataset to join
        model_version: the model version to use from the model artefacts
        model_artefacts_table: the model artefacts table where the model artefacts are stored (defaults to vmo_sto_models)
        partition_id: the dataset partition_id

    Returns:
        DataFrame
    """
    query = f"""
    SELECT d.*, CASE WHEN n_row=1 THEN m.model ELSE null END AS model 
        FROM (SEL x.*, ROW_NUMBER() OVER (PARTITION BY x.{partition_id} ORDER BY x.{partition_id}) AS n_row FROM {data_table} x) AS d
        CROSS JOIN (SEL * FROM {model_artefacts_table} WHERE model_version = '{model_version}' QUALIFY CSUM(1,1) = 1) m
    """

    return DataFrame.from_query(query)


def get_joined_partitioned_models_df(
    data_table: str,
    model_version: str,
    model_artefacts_table: str = DEFAULT_STO_MODELS_TABLE,
    partition_id: str = "partition_id",
):
    """
    Joins the dataset which is to be used for scoring/evaluation with the model artefacts and appends the artefact
    to the first row with the column name 'model'.

    Args:
        data_table: the table/view of the dataset to join
        model_version: the model version to use from the model artefacts
        model_artefacts_table: the model artefacts table where the model artefacts are stored (defaults to vmo_sto_models)
        partition_id: the dataset partition_id

    Returns:
        DataFrame
    """
    return get_joined_models_df(
        data_table, model_artefacts_table, model_version, partition_id
    )



import json
import numpy as np
import pandas as pd
import math
import logging

from typing import List, Dict
from teradataml.analytics.valib import *
from teradataml import DataFrame
from decimal import Decimal
from aoa.stats import store

logger = logging.getLogger(__name__)

value_for_nan = 0


class _NpEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating) or isinstance(obj, Decimal):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return super(_NpEncoder, self).default(obj)


def _compute_continuous_edges(
    variables: List[str], statistics: pd.DataFrame, dtypes: Dict, bins=10
) -> List[float]:
    edges = []

    # should return what is in the feature catalog.. for now calculate linspace boundaries based on min/max
    ranges = statistics.drop(
        statistics.columns.difference(["xcol", "xmin", "xmax"]), axis=1
    )
    ranges = ranges.set_index("xcol")
    ranges.index = ranges.index.map(str.lower)
    ranges = ranges.to_dict(orient="index")

    for var in variables:
        x_min, x_max = ranges[var]["xmin"], ranges[var]["xmax"]

        # if integer type and range is less than the number of bins, only use 'range' bins
        if x_max - x_min < bins and dtypes[var].startswith("int"):
            edges.append(
                np.linspace(x_min, x_max, int(x_max) - int(x_min) + 1).tolist()
            )
        elif dtypes[var].startswith("decimal") or dtypes[var].startswith("float"):
            # (min / max must be rounded up / down accordingly so easier to just use the vals from stats)
            vals = np.linspace(x_min, x_max, bins + 1).tolist()

            # Rounding all edges to the best Teradata-compatible precision
            vals = list(map(_round_to_td_float, vals))

            dedup = []
            [dedup.append(x) for x in vals if x not in dedup]

            edges.append(dedup)
        else:  # Here we compute for regular int variables
            vals = np.linspace(x_min, x_max, bins + 1).tolist()
            for i in range(1, len(vals)):
                vals[i] = round(vals[i], 1)
            edges.append(vals)

    return edges


def _convert_all_edges_to_val_str(all_edges):
    # boundaries for multiple columns follows the following format..
    # ["{10, 0, 200000}", "{5, 0, 100}"]
    boundaries = []
    for edges in all_edges:
        edges_str = ",".join(str(edge) for edge in edges)
        boundaries.append("{{ {} }}".format(edges_str))

    return boundaries


def _fill_missing_bins(
    bin_edges: List[float], bin_values: List[float], reference_edges: List[float]
) -> List[float]:
    """
    Compare the `bin_edges` returned by VAL to `reference_edges` and fill in any missing bins with `0`.
    This is required as VAL doesn't return empty bins and so if we want to ensure we always have all the bins
    represented in the `bin_values` (which we do or else indices vs reference edges mean nothing), then we must do this.

    """
    new_bin_values = list(bin_values)

    epsilon = 1e-08
    for i, edge in enumerate(reference_edges):
        is_present = False
        for curr_edge in bin_edges:
            if abs(float(curr_edge) - float(edge)) < epsilon:
                is_present = True

        if not is_present:
            new_bin_values.insert(i, 0.0)

    return new_bin_values


def _process_categorical_var(
    frequencies: pd.DataFrame,
    group_label: str,
    variable_name: str,
    feature_importance: float,
    reference_categories: List,
) -> Dict:
    """
    Adds the following struct
    {
        "type": "categorical",
        "group": "<group-name>",
        "importance": "<feature-importance>",
        "statistics": {
            "nulls": <missing-values>,
            "categories": ["cat1", "cat2", ...]
            "frequency": {
                "cat1": 50,
                "cat2": 10,
                ...
            }
        }
    }
    """

    data_struct = {"type": "categorical", "group": group_label, "statistics": {}}

    if feature_importance:
        data_struct["importance"] = feature_importance

    var_freq = frequencies[frequencies.xcol == variable_name]

    # if first row is nan then it is the null values in the dataset. remove from histogram
    if var_freq["xval"].isnull().values.any():
        n = var_freq[var_freq["xval"].isnull()]
        data_struct["statistics"]["nulls"] = n.xcnt.tolist()[0]
        var_freq = var_freq[var_freq["xval"].notnull()]

    if var_freq.empty:
        # We should probably generate all-zero histogram here?
        logger.warning(
            f"Categorical feature {variable_name} has only NULL values, not computing a"
            " frequency"
        )
        return data_struct

    frequencies_dict = (
        var_freq[["xval", "xcnt"]].set_index("xval").T.to_dict(orient="records")[0]
    )

    if reference_categories is None:
        # No reference categories - we stop here, not reporting frequencies
        data_struct["statistics"]["unmonitored_frequency"] = frequencies_dict
        return data_struct
    else:

        # Record the reference categories. This is so we can map based on category -> index consistently for the
        # prometheus labels. This should be removed when this mapping is removed from handling in prometheus on backend.
        data_struct["categories"] = reference_categories

        # Record the `statistics.frequency` for each category (using the category as key).
        # `0` is set as the frequency for any category which is in `reference_categories` but not in `frequencies_dict`
        # This approach allows for new categories to be added over time (never removed) to reference_categories and
        # for everything to remain consistent.
        data_struct["statistics"]["frequency"] = {
            cat: frequencies_dict.get(cat, 0) if cat in reference_categories else 0
            for i, cat in enumerate(reference_categories)
        }
        for f in frequencies_dict.keys():
            if f not in reference_categories:
                logger.warning(
                    f"Categorical feature {variable_name} has a new category {f} that's"
                    " not listed in reference categories"
                )
                if "unmonitored_frequency" not in data_struct["statistics"]:
                    data_struct["statistics"]["unmonitored_frequency"] = {}
                data_struct["statistics"]["unmonitored_frequency"][f] = (
                    frequencies_dict.get(f)
                )

        return data_struct


def _process_continuous_var(
    histogram: pd.DataFrame,
    stats: pd.DataFrame,
    group_label: str,
    variable_name: str,
    feature_importance: float,
    reference_edges: List[float],
) -> Dict:
    """
    Adds the following struct
    {
        "type": "continuous",
        "group": "<group-name>",
        "importance": "<feature-importance>",
        "statistics": {
            "nulls": <missing-values>,
            "cnt": 614.0,
            "min": 21.0,
            "max": 81.0,
            ...
            "histogram": {
              "edges": [21.0,27.0,33.0,39.0,45.0,51.0,57.0,63.0,69.0,75.0,81.0],
              "values": [248.0,122.0,77.0,62.0,45.0,23.0,20.0,13.0,3.0,1.0]
            }
        }
    }
    """

    data_struct = {
        "type": "continuous",
        "group": group_label,
        "statistics": {},
    }

    if feature_importance:
        data_struct["importance"] = feature_importance

    def _strip_key_x(d: Dict):
        return {k[1:]: v for k, v in d.items()}

    stats_values = (
        stats[stats.xcol == variable_name]
        .drop(["xdb", "xtbl", "xcol"], axis=1)
        .to_dict(orient="records")[0]
    )
    data_struct["statistics"].update(_strip_key_x(stats_values))

    if histogram.empty:
        # No histogram available - we stop here, not reporting histogram
        return data_struct

    var_hist = histogram[histogram.xcol == variable_name].sort_values(by=["xbin"])

    # if first row is nan then it is the null values in the dataset. remove from histogram
    if var_hist["xbin"].isnull().values.any():
        n = var_hist[var_hist["xbin"].isnull()]
        data_struct["statistics"]["nulls"] = n.xcnt.tolist()[0]

        var_hist = var_hist[var_hist["xbin"].notnull()]

    if var_hist.empty:
        # We should probably generate all-zero histogram here?
        logger.warning(
            f"Continuous feature {variable_name} has only NULL values, not computing a"
            " histogram"
        )
        return data_struct

    bin_edges = [var_hist.xbeg.tolist()[0]] + var_hist.xend.tolist()
    bin_values = var_hist.xcnt.tolist()

    # (issue #123) VAL docs originally stated that:
    # VAL histograms will values lower than the first bin to the first bin, but values greater than the
    # largest bin are added to a new bin.. Therefore we did the same on both sides. However, it turns out this doc is
    # incorrect.

    is_right_outlier_bin = math.isnan(bin_edges[-1])
    is_left_outlier_bin = math.isnan(bin_edges[0])
    if is_right_outlier_bin:
        bin_edges = bin_edges[:-1]
    if is_left_outlier_bin:
        bin_edges = bin_edges[1:]

    # Add missing bin_values based on the bin_edges vs reference_edges.
    # VAL doesn't return empty bins
    if len(bin_edges) < len(reference_edges):
        bin_values = _fill_missing_bins(
            bin_edges=bin_edges, bin_values=bin_values, reference_edges=reference_edges
        )

    right_outliers = 0
    left_outliers = 0
    if is_right_outlier_bin:
        right_outliers = int(bin_values[-1])
        logger.warning(
            f"Continuous feature {variable_name} has values larger than the rightmost"
            f" bin, identified {right_outliers} values"
        )
        bin_values[-2] += bin_values[
            -1
        ]  # Probably should consider removing as some point, when outliers are properly reported and visible to the user
        bin_values = bin_values[:-1]
    if is_left_outlier_bin:
        left_outliers = int(bin_values[0])
        logger.warning(
            f"Continuous feature {variable_name} has values smaller than the leftmost"
            f" bin, identified {left_outliers} values"
        )
        bin_values[1] += bin_values[
            0
        ]  # Probably should consider removing as some point, when outliers are properly reported and visible to the user
        bin_values = bin_values[1:]

    data_struct["statistics"]["histogram"] = {
        "edges": reference_edges,
        "values": bin_values,
    }

    if is_right_outlier_bin:
        data_struct["statistics"]["histogram"]["right_outliers"] = right_outliers
    if is_left_outlier_bin:
        data_struct["statistics"]["histogram"]["left_outliers"] = left_outliers

    return data_struct


def _parse_scoring_stats(
    features_df: DataFrame,
    predicted_df: DataFrame,
    data_stats: Dict,
    feature_importance: Dict[str, float] = {},
    feature_metadata_fqtn: str = None,
    feature_metadata_group: str = None,
) -> Dict:
    if not isinstance(features_df, DataFrame):
        raise TypeError("We only support teradataml DataFrame for features")

    if not isinstance(predicted_df, DataFrame):
        raise TypeError("We only support teradataml DataFrame for predictions")

    features = []
    targets = []
    categorical = []

    # ensure backward compatible (when we had targets incorrectly named as predictors..)
    if "predictors" in data_stats:
        data_stats["targets"] = data_stats.pop("predictors")

    for var_type in ["features", "targets"]:
        for name, value in data_stats[var_type].items():
            # for backward compatibility with data stats created before we lower-cased
            name = name.lower()

            if var_type == "features":
                features.append(name)
            elif var_type == "targets":
                targets.append(name)
            if "type" in value and value["type"] == "categorical":
                categorical.append(name)

    if predicted_df.shape[0] != features_df.shape[0]:
        raise ValueError(
            "The number of prediction rows do not match the number of features rows!"
        )

    data_stats = _capture_stats(
        df=features_df,
        features=features,
        targets=[],
        categorical=categorical,
        feature_importance=feature_importance,
        feature_metadata_fqtn=feature_metadata_fqtn,
        feature_metadata_group=feature_metadata_group,
    )

    targets_struct = _capture_stats(
        df=predicted_df,
        features=[],
        targets=targets,
        categorical=categorical,
        feature_importance=feature_importance,
        feature_metadata_fqtn=feature_metadata_fqtn,
        feature_metadata_group=feature_metadata_group,
    )

    data_stats["targets"] = targets_struct["targets"]

    return data_stats


def _capture_stats(
    df: DataFrame,
    features: List,
    targets: List,
    categorical: List,
    feature_importance: Dict[str, float] = {},
    feature_metadata_fqtn: str = {},
    feature_metadata_group: str = "default",
) -> Dict:
    if not isinstance(df, DataFrame):
        raise TypeError("We only support teradataml DataFrame")

    # lowercase all keys/names to avoid mismatches between sql/dataframes case sensitivity
    features = [f.lower() for f in features]
    targets = [t.lower() for t in targets]
    categorical = [c.lower() for c in categorical]
    feature_importance = {k.lower(): v for k, v in feature_importance.items()}
    df_columns = [c.lower() for c in df.columns]

    # validate that the dataframe contains the features/targets provided
    if features and not set(features).issubset(df_columns):
        raise ValueError(
            f"features dataframe with columns ({df.columns}) does not contain features:"
            f" {features}"
        )

    if targets and not set(targets).issubset(df_columns):
        raise ValueError(
            f"targets dataframe with columns ({df.columns}) does not contain targets:"
            f" {targets}"
        )

    if not feature_metadata_fqtn:
        raise ValueError("feature_metadata_fqtn must be defined")

    total_rows = df.shape[0]
    continuous_vars = list((set(features) | set(targets)) - set(categorical))
    categorical_vars = list((set(features) | set(targets)) - set(continuous_vars))
    reference_edges = []
    reference_categories = []
    available_continuous_vars = []
    available_categorical_vars = []

    if len(continuous_vars) > 0:
        logger.debug(
            "Executing this VAL"
            f" {valib.Statistics(data=df, columns=continuous_vars, stats_options='all').show_query()}"
        )
        stats = valib.Statistics(data=df, columns=continuous_vars, stats_options="all")
        stats = stats.result.to_pandas().fillna(value_for_nan).reset_index()
        stats["xcol"] = stats["xcol"].str.lower()

        stats_metadata = store.get_feature_stats(feature_metadata_fqtn, "continuous")

        for v in continuous_vars:
            if v.lower() in stats_metadata.keys():
                reference_edges.append(stats_metadata[v.lower()]["edges"])
                available_continuous_vars.append(v)
            else:
                logger.warning(
                    f"Feature {v} doesn't have statistics metadata defined, and will"
                    " not be monitored.\nIn order to enable monitoring for this"
                    " feature, make sure that statistics metadata is availabe in"
                    f" {feature_metadata_fqtn}"
                )

        if len(available_continuous_vars) > 0:
            hist_boundaries = _convert_all_edges_to_val_str(reference_edges)
            histogram = _histogram_wrapper(
                data=df, columns=available_continuous_vars, boundaries=hist_boundaries
            )
            histogram["xcol"] = histogram["xcol"].str.lower()

    if len(categorical_vars) > 0:
        stats_metadata = store.get_feature_stats(feature_metadata_fqtn, "categorical")

        for v in categorical_vars:
            if v.lower() in stats_metadata.keys():
                reference_categories.append(stats_metadata[v.lower()]["categories"])
                available_categorical_vars.append(v)
            else:
                logger.warning(
                    f"Feature {v} doesn't have statistics metadata defined, and will"
                    " not be monitored.\nIn order to enable monitoring for this"
                    " feature, make sure that statistics metadata is availabe in"
                    f" {feature_metadata_fqtn}"
                )

        # We compute frequencies for all categorical features independent of metadata availability
        # But we only reporing number of nulls if reference categories are not available
        logger.debug(
            "Executing this VAL"
            f" {valib.Frequency(data=df, columns=categorical_vars).show_query()}"
        )
        frequencies = valib.Frequency(data=df, columns=categorical_vars)
        frequencies = frequencies.result.to_pandas(all_rows=True).reset_index()
        frequencies["xcol"] = frequencies["xcol"].str.lower()

    data_struct = {"num_rows": total_rows, "features": {}, "targets": {}}

    def add_var_metadata(variable_type, group_label):
        if variable_name in continuous_vars:
            if variable_name in available_continuous_vars:
                data_struct[variable_type][variable_name] = _process_continuous_var(
                    histogram=histogram,
                    stats=stats,
                    reference_edges=reference_edges[
                        available_continuous_vars.index(variable_name)
                    ],
                    group_label=group_label,
                    variable_name=variable_name,
                    feature_importance=feature_importance.get(variable_name, None),
                )
            else:
                data_struct[variable_type][variable_name] = _process_continuous_var(
                    histogram=pd.DataFrame(),
                    stats=stats,
                    reference_edges=None,
                    group_label=group_label,
                    variable_name=variable_name,
                    feature_importance=feature_importance.get(variable_name, None),
                )
        else:
            if variable_name in available_categorical_vars:
                data_struct[variable_type][variable_name] = _process_categorical_var(
                    frequencies=frequencies,
                    group_label=group_label,
                    variable_name=variable_name,
                    feature_importance=feature_importance.get(variable_name, None),
                    reference_categories=reference_categories[
                        available_categorical_vars.index(variable_name)
                    ],
                )
            else:
                data_struct[variable_type][variable_name] = _process_categorical_var(
                    frequencies=frequencies,
                    group_label=group_label,
                    variable_name=variable_name,
                    feature_importance=feature_importance.get(variable_name, None),
                    reference_categories=None,
                )

    for variable_name in features:
        add_var_metadata("features", feature_metadata_group)

    for variable_name in targets:
        add_var_metadata("targets", feature_metadata_group)

    missing_continuous = [
        c for c in continuous_vars if not (c.lower() in available_continuous_vars)
    ]
    missing_categorical = [
        c for c in categorical_vars if not (c.lower() in available_categorical_vars)
    ]

    if len(missing_categorical) > 0 or len(missing_continuous) > 0:
        data_struct["missing_metadata"] = {}
        if len(missing_continuous) > 0:
            data_struct["missing_metadata"]["continuous"] = missing_continuous
        if len(missing_categorical) > 0:
            data_struct["missing_metadata"]["categorical"] = missing_categorical

    return data_struct


def _histogram_wrapper(data, columns, boundaries):
    working_columns = list(columns)
    working_boundaries = list(boundaries)
    result = []

    # Iterate over an updated list of columns while it's not empty
    while working_columns:
        tc = list(working_columns)
        tb = list(working_boundaries)
        # If the size of td_analyze parameter is bigger than the limt
        # then reduce the list of columns until it fits
        while _val_parameter_length_too_long(tc, tb):
            if len(tc) == 1:
                raise Exception(
                    "VAL histogram parameters are too long even for single column"
                    f" {tc[0]}"
                )
            tc.pop()
            tb.pop()

        # compute a histogram on the longest possible list of columns
        logger.debug(
            "Executing this VAL"
            f" {valib.Histogram(data=data,columns=tc,boundaries=tb).show_query()}"
        )
        histogram = valib.Histogram(data=data, columns=tc, boundaries=tb)
        result.append(histogram.result.to_pandas(all_rows=True).reset_index())

        # removing columns thats already computed
        working_columns = working_columns[len(tc) :]
        working_boundaries = working_boundaries[len(tb) :]

    return pd.concat(result)


def _val_parameter_length_too_long(columns, boundaries, limit=31000):
    db_tbl_string = "database=;tablename=;outputdatabase=;outputtablename=;"
    column_string = f"columns={','.join(columns)};boundaries={','.join(boundaries)};"
    # assuming maximum length for database and table names at 128
    final_length = len(db_tbl_string) + len(column_string) + 128 * 4
    if final_length > limit:
        logger.debug(
            "VAL histogram parameters are too long for these columns, trying to"
            f" compute sequentially: {','.join(columns)}"
        )
    return final_length > limit


def _round_to_td_float(x, max_digits=15):
    # This function is needed to cast/round decimal and float number to the maximum precision allowed by Teradata database.
    # Teradata floats allow maximum 15 digits for mantissa, so this strange logic below does exactly that.
    # Possibly at some point we may need to add a round to exponent as well.
    (sign, digits, exponent) = Decimal(x).as_tuple()
    e = len(digits) + exponent
    (_, m, _) = (
        Decimal(x)
        .scaleb(-e)
        .quantize(Decimal(f"1E-{max_digits}"))
        .normalize()
        .as_tuple()
    )
    return np.float64(Decimal((sign, m, e - len(m))))



from __future__ import absolute_import

from aoa.api.iterator_base_api import IteratorBaseApi


class JobApi(IteratorBaseApi):
    path = "/api/jobs"
    type = "JOB"

    def find_by_archived(
        self,
        archived: bool = False,
        projection: str = None,
        page: int = None,
        size: int = None,
        sort: str = None,
    ):
        raise NotImplementedError("Archiving not supported for Jobs")

    def _get_header_params(self):
        header_vars = [
            "AOA-Project-ID",
            "VMO-Project-ID",
            "Content-Type",
            "Accept",
        ]  # AOA-Project-ID kept for backwards compatibility
        header_vals = [
            self.aoa_client.project_id,
            self.aoa_client.project_id,
            self.json_type,
            self.aoa_client.select_header_accept([
                self.json_type,
                "application/hal+json",
                "text/uri-list",
                "application/x-spring-data-compact+json",
            ]),
        ]

        return self.generate_params(header_vars, header_vals)

    def find_job_events(self, job_id: str, projection: str = None):
        """
        returns events of a job

        Parameters:
           job_id (str): job id(uuid) to find events
           projection (str): projection type

        Returns:
            (dict): job events
        """
        query_vars = ["projection"]
        query_vals = [projection]
        query_params = self.generate_params(query_vars, query_vals)

        return self.aoa_client.get_request(
            path=f"{self.path}/{job_id}/events",
            header_params=self._get_header_params(),
            query_params=query_params,
        )

    def find_by_job_event_id(
        self, job_id: str, job_event_id: str, projection: str = None
    ):
        """
        returns job event

        Parameters:
           job_id (str): job id(uuid)
           job_event_id (str): job event id(uuid)
           projection (str): projection type

        Returns:
            (dict): job event
        """
        query_vars = ["projection"]
        query_vals = [projection]
        query_params = self.generate_params(query_vars, query_vals)

        return self.aoa_client.get_request(
            path=f"{self.path}/{job_id}/events/{job_event_id}",
            header_params=self._get_header_params(),
            query_params=query_params,
        )

    def find_job_between_intervals(
        self, start_time: str, end_time: str, projection: str = None
    ):
        """
        returns job events within the timestamp range

        Parameters:
           start_time (str): start time e.g: 2020-03-18T09:05:03.569Z
           end_time (str): end time e.g: 2020-03-24T09:05:03.569Z
           projection (str): projection type

        Returns:
            (dict): job events
        """
        query_vars = ["startAt", "endAt", "projection"]
        query_vals = [start_time, end_time, projection]
        query_params = self.generate_params(query_vars, query_vals)

        return self.aoa_client.get_request(
            path=f"{self.path}/search/findByCreatedAtBetween",
            header_params=self._get_header_params(),
            query_params=query_params,
        )

    def find_by_status(
        self,
        status: str,
        projection: str = None,
        page: int = None,
        size: int = None,
        sort: str = None,
    ):
        """
        returns job by status

        Parameters:
           status (str): job status
           projection (str): projection type
           page (int): page number
           size (int): number of records in a page
           sort (str): column name and sorting order
           e.g. name?asc: sort name in ascending order, name?desc: sort name in descending order

        Returns:
            (dict): job events
        """
        query_vars = ["status", "projection", "page", "sort", "size"]
        query_vals = [status, projection, page, sort, size]
        query_params = self.generate_params(query_vars, query_vals)

        return self.aoa_client.get_request(
            path=f"{self.path}/search/findByStatusIn",
            header_params=self._get_header_params(),
            query_params=query_params,
        )

    def wait(self, job_id: str, timeout_sec: int = 60):
        import time

        start_time_sec = int(time.time())

        while True:
            job = self.find_by_id(id=job_id, projection="expandJob")
            status = job.get("status", "UNKNOWN")

            if status == "COMPLETED":
                return
            elif status in ["ERROR", "CANCELLED"]:
                raise SystemError(f"Job failed with status: {status}")

            if int(start_time_sec) - start_time_sec > timeout_sec:
                raise TimeoutError(
                    f"Timeout waiting for job to complete. Current status: {status}"
                )

            time.sleep(5)



from __future__ import absolute_import

from aoa.api.iterator_base_api import IteratorBaseApi


class DeploymentApi(IteratorBaseApi):
    path = "/api/deployments"
    type = "DEPLOYMENT"

    def _get_header_params(self):
        header_vars = [
            "AOA-Project-ID",
            "VMO-Project-ID",
            "Content-Type",
            "Accept",
        ]  # AOA-Project-ID kept for backwards compatibility
        header_vals = [
            self.aoa_client.project_id,
            self.aoa_client.project_id,
            self.json_type,
            self.aoa_client.select_header_accept([
                self.json_type,
                "application/hal+json",
                "text/uri-list",
                "application/x-spring-data-compact+json",
            ]),
        ]

        return self.generate_params(header_vars, header_vals)

    def find_by_archived(
        self,
        archived: bool = False,
        projection: str = None,
        page: int = None,
        size: int = None,
        sort: str = None,
    ):
        raise NotImplementedError("Archiving not supported for Deployments")

    def find_active_by_trained_model_and_engine_type(
        self, trained_model_id: str, engine_type: str, projection: str = None
    ):
        """
        returns deployments by trained model and engine type

        Parameters:
           trained_model_id (str): trained model id(string) to find
           engine_type (str): engine type(string) to find
           projection (str): projection type

        Returns:
            (dict): deployments
        """
        query_vars = ["trainedModelId", "engineType", "projection"]
        query_vals = [trained_model_id, engine_type, projection]
        query_params = self.generate_params(query_vars, query_vals)

        return self.aoa_client.get_request(
            path=f"{self.path}/search/findActiveByTrainedModelIdAndEngineType",
            header_params=self._get_header_params(),
            query_params=query_params,
        )

    def find_by_status(self, status: str, projection: str = None):
        """
        returns deployments by status
        Parameters:
           status (str): status(string) to find
           projection (str): projection type
        Returns:
            (dict): deployments
        """
        query_vars = ["status", "projection"]
        query_vals = [status, projection]
        query_params = self.generate_params(query_vars, query_vals)

        return self.aoa_client.get_request(
            path=f"{self.path}/search/findByStatus",
            header_params=self._get_header_params(),
            query_params=query_params,
        )

    def find_active(self, projection: str = None):
        """
        returns active deployments
        Parameters:
           projection (str): projection type
        Returns:
            (dict): deployments
        """
        query_vars = ["projection"]
        query_vals = [projection]
        query_params = self.generate_params(query_vars, query_vals)

        return self.aoa_client.get_request(
            path=f"{self.path}/search/findActive",
            header_params=self._get_header_params(),
            query_params=query_params,
        )

    def find_active_by_task_id(self, task_id: str = None, projection: str = None):
        """
        returns active deployments
        Parameters:
           projection (str): projection type
        Returns:
            (dict): deployments
        """
        query_vars = ["taskId", "projection"]
        query_vals = [task_id, projection]
        query_params = self.generate_params(query_vars, query_vals)

        return self.aoa_client.get_request(
            path=f"{self.path}/search/findActiveByTaskId",
            header_params=self._get_header_params(),
            query_params=query_params,
        )



from __future__ import absolute_import
from typing import Dict

from aoa.api.iterator_base_api import IteratorBaseApi


class DatasetConnectionApi(IteratorBaseApi):
    path = "/api/datasetConnections"
    type = "DATASET_CONNECTION"

    def find_by_archived(
        self,
        archived: bool = False,
        projection: str = None,
        page: int = None,
        size: int = None,
        sort: str = None,
    ):
        raise NotImplementedError("Archiving not supported for DatasetConnections")

    def _get_header_params(self):
        header_vars = [
            "AOA-Project-ID",
            "VMO-Project-ID",
            "Content-Type",
            "Accept",
        ]  # AOA-Project-ID kept for backwards compatibility
        header_vals = [
            self.aoa_client.project_id,
            self.aoa_client.project_id,
            self.json_type,
            self.aoa_client.select_header_accept([
                self.json_type,
                "application/hal+json",
                "text/uri-list",
                "application/x-spring-data-compact+json",
            ]),
        ]

        return self.generate_params(header_vars, header_vals)

    def save(self, dataset_connection: Dict[str, str]):
        """
        register a dataset connection

        Parameters:
           dataset connection (dict): dataset connection to register

        Returns:
            (dict): dataset template
        """

        return self.aoa_client.post_request(
            path=self.path,
            header_params=self._get_header_params(),
            query_params={},
            body=dataset_connection,
        )



from teradataml import DataFrame, copy_to_sql
from aoa import aoa_create_context, ModelContext
import pandas as pd
import numpy as np


def run_task(context: ModelContext, **kwargs):
    aoa_create_context()
    df = DataFrame.from_query("sel * from pima_patient_features")

    # Convert teradataml DataFrame to pandas DataFrame
    df_pd = df.to_pandas()

    # Generate random birthdates
    start_date = pd.to_datetime("1950-01-01")
    end_date = pd.to_datetime("2000-01-01")

    df_pd["birthday"] = start_date + (end_date - start_date) * np.random.rand(
        len(df_pd)
    )

    # Calculate age
    df_pd["calculated_age"] = df_pd["birthday"].apply(
        lambda x: (pd.to_datetime("today") - x).days // 365
    )

    # Remove the original age column
    df_pd = df_pd.drop(columns=["Age"])

    # Write pandas DataFrame to a Teradata table
    copy_to_sql(df=df_pd, table_name="age_table", if_exists="replace")

    # Create a teradataml DataFrame from the table
    df = DataFrame("age_table")

    print(df)
    with open(f"{context.artifact_output_path}/age_report.txt", "w") as f:
        print(df, file=f)

    # Store build properties as a file artifact
    with open(f"{context.artifact_output_path}/build_properties.txt", "w") as f:
        f.write(str(kwargs))



from __future__ import absolute_import


class ApiIterator(object):
    _pos = 0
    _current_page = 0

    def __init__(
        self,
        api_func=None,
        entities: str = None,
        projection: str = None,
        page_size: int = None,
        sort: str = None,
    ):
        if api_func is None or entities is None:
            pass
        else:
            self._api_func = api_func
            self._projection = projection
            self._page_size = page_size
            self._sort = sort
            page_info = api_func(
                projection=projection, page=0, size=page_size, sort=sort
            )["page"]
            self._total_elements = page_info["totalElements"]
            self._total_pages = page_info["totalPages"]
            self._cache = api_func(
                projection=projection, page=0, size=page_size, sort=sort
            )["_embedded"][entities]
            self._entities = entities

    def __next__(self):
        if self._pos < len(self._cache):
            result = self._cache[self._pos]
            self._pos += 1
            return result
        elif self._current_page < self._total_pages - 1:
            self._current_page += 1
            self._pos = 1  # so we skip the first one in next iteration
            self._cache = self._api_func(
                projection=self._projection,
                page=self._current_page,
                size=self._page_size,
                sort=self._sort,
            )["_embedded"][self._entities]
            return self._cache[0]
        else:
            raise StopIteration

    def __iter__(self):
        return self



from __future__ import absolute_import
from typing import Dict, List

from aoa.api.iterator_base_api import IteratorBaseApi


class ProjectApi(IteratorBaseApi):

    path = "/api/projects"
    type = "PROJECT"

    def _get_header_params(self):
        # The header for project id is required for the archive/unarchive method from base_api
        header_vars = [
            "AOA-Project-ID",
            "VMO-Project-ID",
            "Content-Type",
            "Accept",
        ]  # AOA-Project-ID kept for backwards compatibility
        header_vals = [
            self.aoa_client.project_id,
            self.aoa_client.project_id,
            self.json_type,
            self.json_type,
        ]

        return self.generate_params(header_vars, header_vals)

    def find_by_name_like(self, name: str, projection: str = None) -> List:
        """
        returns a list of projects matching the given name

        Parameters:
           name (str): project name to search by
           projection (str): projection type

        Returns:
            (list): list of projects
        """
        query_vars = ["name", "projection"]
        query_vals = [name, projection]
        query_params = self.generate_params(query_vars, query_vals)

        return self.aoa_client.get_request(
            path=f"{self.path}/search/findByName",
            header_params=self._get_header_params(),
            query_params=query_params,
        )

    def save(self, project: Dict[str, str]):
        """
        create a project

        Parameters:
           project (dict): project to create

        Returns:
            (dict): project
        """
        header_vars = ["Accept"]
        header_vals = [self.json_type]
        header_params = self.generate_params(header_vars, header_vals)

        return self.aoa_client.post_request(
            path=self.path, header_params=header_params, query_params={}, body=project
        )

    def update(self, project: Dict[str, str]):
        """
        update a project

        Parameters:
           project (dict): project to update

        Returns:
            (dict): project
        """
        header_vars = ["Accept"]
        header_vals = [self.json_type]
        header_params = self.generate_params(header_vars, header_vals)

        return self.aoa_client.put_request(
            path=f"{self.path}/{self.aoa_client.project_id}",
            header_params=header_params,
            query_params={},
            body=project,
        )

    def is_archived(self, project_id: str):
        """
        returns the archived status of a project
        Parameters:
           project_id (str): project id to check archived status
        Returns:
            (bool)
        """
        project = self.find_by_id(project_id)

        return project["archived"]



# Copyright 2018 Teradata. All rights reserved.
#
#   File:       TJEncryptPassword.py
#   Purpose:    Encrypts a password, saves the encryption key in one file, and saves the encrypted password in a second file
#
#               This program accepts eight command-line arguments:
#
#                 1. Transformation - Specifies the transformation in the form Algorithm/Mode/Padding.
#                                     Example: AES/CBC/NoPadding
#
#                 2. KeySizeInBits  - Specifies the algorithm key size, which governs the encryption strength.
#                                     Example: 256
#
#                 3. MAC            - Specifies the message authentication code (MAC) algorithm HmacSHA1 or HmacSHA256.
#                                     Example: HmacSHA256
#
#                 4. PasswordEncryptionKeyFileName - Specifies a filename in the current directory, a relative pathname, or an absolute pathname.
#                                     The file is created by this program. If the file already exists, it will be overwritten by the new file.
#                                     Example: PassKey.properties
#
#                 5. EncryptedPasswordFileName - Specifies a filename in the current directory, a relative pathname, or an absolute pathname.
#                                     The filename or pathname that must differ from the PasswordEncryptionKeyFileName.
#                                     The file is created by this program. If the file already exists, it will be overwritten by the new file.
#                                     Example: EncPass.properties
#
#                 6. Hostname       - Specifies the Teradata Database hostname.
#                                     Example: whomooz
#
#                 7. Username       - Specifies the Teradata Database username.
#                                     Example: guest
#
#                 8. Password       - Specifies the Teradata Database password to be encrypted.
#                                     Unicode characters in the password can be specified with the backslash uXXXX escape sequence.
#                                     Example: please
#
#               Overview
#               --------
#
#               Stored Password Protection enables an application to provide a connection password in encrypted form to
#               the Teradata SQL Driver for Python.
#
#               An encrypted password may be specified in the following contexts:
#                 * A login password specified as the "password" connection parameter.
#                 * A login password specified within the "logdata" connection parameter.
#
#               If the password, however specified, begins with the prefix "ENCRYPTED_PASSWORD(" then the specified password must follow this format:
#
#                 ENCRYPTED_PASSWORD(file:PasswordEncryptionKeyFileName,file:EncryptedPasswordFileName)
#
#               Each filename must be preceded by the "file:"" prefix.
#               The PasswordEncryptionKeyFileName must be separated from the EncryptedPasswordFileName by a single comma.
#               The PasswordEncryptionKeyFileName specifies the name of a properties file that contains the password encryption key and associated information.
#               The EncryptedPasswordFileName specifies the name of a properties file that contains the encrypted password and associated information.
#               The two files are described below.
#
#               Stored Password Protection is offered by the Teradata JDBC Driver and the Teradata SQL Driver for Python.
#               The same file format is used by both drivers.
#
#               This program works in conjunction with Stored Password Protection offered by the Teradata JDBC Driver and the
#               Teradata SQL Driver for Python. This program creates the files containing the password encryption key and encrypted password,
#               which can be subsequently specified via the "ENCRYPTED_PASSWORD(" syntax.
#
#               You are not required to use this program to create the files containing the password encryption key and encrypted password.
#               You can develop your own software to create the necessary files.
#               The only requirement is that the files must match the format expected by the Teradata SQL Driver for Python, which is documented below.
#
#               This program encrypts the password and then immediately decrypts the password, in order to verify that the password can be
#               successfully decrypted. This program mimics the password decryption of the Teradata SQL Driver for Python, and is intended
#               to openly illustrate its operation and enable scrutiny by the community.
#
#               The encrypted password is only as safe as the two files. You are responsible for restricting access to the files containing
#               the password encryption key and encrypted password. If an attacker obtains both files, the password can be decrypted.
#               The operating system file permissions for the two files should be as limited and restrictive as possible, to ensure that
#               only the intended operating system userid has access to the files.
#
#               The two files can be kept on separate physical volumes, to reduce the risk that both files might be lost at the same time.
#               If either or both of the files are located on a network volume, then an encrypted wire protocol can be used to access the
#               network volume, such as sshfs, encrypted NFSv4, or encrypted SMB 3.0.
#
#               Example Commands
#               ----------------
#
#               This program uses the Teradata SQL Driver for Python to log on to the specified Teradata Database using the encrypted
#               password, so the Teradata SQL Driver for Python must have been installed with the "pip install teradatasql" command.
#
#               The following commands assume this program file is located in the current directory.
#               When the Teradata SQL Driver for Python is installed, the sample programs are placed in the teradatasql/samples
#               directory under your Python installation directory.
#               Change your current directory to the teradatasql/samples directory under your Python installation directory.
#
#               The following example commands illustrate using a 256-bit AES key, and using the HmacSHA256 algorithm.
#
#               macOS or Linux: python TJEncryptPassword.py AES/CBC/NoPadding 256 HmacSHA256 PassKey.properties EncPass.properties whomooz guest please
#               Windows:        py -3 TJEncryptPassword.py AES/CBC/NoPadding 256 HmacSHA256 PassKey.properties EncPass.properties whomooz guest please
#
#               Password Encryption Key File Format
#               -----------------------------------
#
#               You are not required to use the TJEncryptPassword program to create the files containing the password encryption key and
#               encrypted password. You can develop your own software to create the necessary files, but the files must match the format
#               expected by the Teradata SQL Driver for Python.
#
#               The password encryption key file is a text file in Java Properties file format, using the ISO 8859-1 character encoding.
#
#               The file must contain the following string properties:
#
#                 version=1
#
#                       The version number must be 1.
#                       This property is required.
#
#                 transformation=TransformationName
#
#                       Specifies the transformation in the form Algorithm/Mode/Padding.
#                       This property is required.
#                       Example: AES/CBC/NoPadding
#
#                 algorithm=AlgorithmName
#
#                       This value must correspond to the Algorithm portion of the transformation.
#                       This property is required.
#                       Example: AES
#
#                 match=MatchValue
#
#                       The password encryption key and encrypted password files must contain the same match value.
#                       The match values are compared to ensure that the two specified files are related to each other,
#                       serving as a "sanity check" to help avoid configuration errors.
#                       This property is required.
#
#                       This program uses a timestamp as a shared match value, but a timestamp is not required.
#                       Any shared string can serve as a match value. The timestamp is not related in any way to
#                       the encryption of the password, and the timestamp cannot be used to decrypt the password.
#
#                 key=HexDigits
#
#                       This value is the password encryption key, encoded as hex digits.
#                       This property is required.
#
#                 mac=MACAlgorithmName
#
#                       Specifies the message authentication code (MAC) algorithm HmacSHA1 or HmacSHA256.
#                       Stored Password Protection performs Encrypt-then-MAC for protection from a padding oracle attack.
#                       This property is required.
#
#                 mackey=HexDigits
#
#                       This value is the MAC key, encoded as hex digits.
#                       This property is required.
#
#               Encrypted Password File Format
#               ------------------------------
#
#               The encrypted password file is a text file in Java Properties file format, using the ISO 8859-1 character encoding.
#
#               The file must contain the following string properties:
#
#                 version=1
#
#                       The version number must be 1.
#                       This property is required.
#
#                 match=MatchValue
#
#                       The password encryption key and encrypted password files must contain the same match value.
#                       The match values are compared to ensure that the two specified files are related to each other,
#                       serving as a "sanity check" to help avoid configuration errors.
#                       This property is required.
#
#                       This program uses a timestamp as a shared match value, but a timestamp is not required.
#                       Any shared string can serve as a match value. The timestamp is not related in any way to
#                       the encryption of the password, and the timestamp cannot be used to decrypt the password.
#
#                 password=HexDigits
#
#                       This value is the encrypted password, encoded as hex digits.
#                       This property is required.
#
#                 params=HexDigits
#
#                       This value contains the cipher algorithm parameters, if any, encoded as hex digits.
#                       Some ciphers need algorithm parameters that cannot be derived from the key, such as an initialization vector.
#                       This property is optional, depending on whether the cipher algorithm has associated parameters.
#
#                       While this value is technically optional, an initialization vector is required by all three
#                       block cipher modes CBC, CFB, and OFB that are supported by the Teradata SQL Driver for Python.
#                       ECB (Electronic Codebook) does not require params, but ECB is not supported by the Teradata SQL Driver for Python.
#
#                 hash=HexDigits
#
#                       This value is the expected message authentication code (MAC), encoded as hex digits.
#                       After encryption, the expected MAC is calculated using the ciphertext, transformation name, and algorithm parameters if any.
#                       Before decryption, the Teradata SQL Driver for Python calculates the MAC using the ciphertext, transformation name,
#                       and algorithm parameters, if any, and verifies that the calculated MAC matches the expected MAC.
#                       If the calculated MAC differs from the expected MAC, then either or both of the files may have been tampered with.
#                       This property is required.
#
#               Transformation, Key Size, and MAC
#               ---------------------------------
#
#               A transformation is a string that describes the set of operations to be performed on the given input, to produce transformed output.
#               A transformation specifies the name of a cryptographic algorithm such as DES or AES, followed by a feedback mode and padding scheme.
#
#               The Teradata SQL Driver for Python supports the following transformations and key sizes.
#
#                  DES/CBC/NoPadding          64
#                  DES/CBC/PKCS5Padding       64
#                  DES/CFB/NoPadding          64
#                  DES/CFB/PKCS5Padding       64
#                  DES/OFB/NoPadding          64
#                  DES/OFB/PKCS5Padding       64
#                  DESede/CBC/NoPadding       192
#                  DESede/CBC/PKCS5Padding    192
#                  DESede/CFB/NoPadding       192
#                  DESede/CFB/PKCS5Padding    192
#                  DESede/OFB/NoPadding       192
#                  DESede/OFB/PKCS5Padding    192
#                  AES/CBC/NoPadding          128
#                  AES/CBC/NoPadding          192
#                  AES/CBC/NoPadding          256
#                  AES/CBC/PKCS5Padding       128
#                  AES/CBC/PKCS5Padding       192
#                  AES/CBC/PKCS5Padding       256
#                  AES/CFB/NoPadding          128
#                  AES/CFB/NoPadding          192
#                  AES/CFB/NoPadding          256
#                  AES/CFB/PKCS5Padding       128
#                  AES/CFB/PKCS5Padding       192
#                  AES/CFB/PKCS5Padding       256
#                  AES/OFB/NoPadding          128
#                  AES/OFB/NoPadding          192
#                  AES/OFB/NoPadding          256
#                  AES/OFB/PKCS5Padding       128
#                  AES/OFB/PKCS5Padding       192
#                  AES/OFB/PKCS5Padding       256
#
#               Stored Password Protection uses a symmetric encryption algorithm such as DES or AES, in which the same secret key is used for
#               encryption and decryption of the password. Stored Password Protection does not use an asymmetric encryption algorithm such as RSA,
#               with separate public and private keys.
#
#               CBC (Cipher Block Chaining) is a block cipher encryption mode. With CBC, each ciphertext block is dependent on all plaintext blocks
#               processed up to that point. CBC is suitable for encrypting data whose total byte count exceeds the algorithm's block size, and is
#               therefore suitable for use with Stored Password Protection.
#
#               Stored Password Protection hides the password length in the encrypted password file by extending the length of the UTF8-encoded
#               password with trailing null bytes. The length is extended to the next 512-byte boundary.
#
#               A block cipher with no padding, such as AES/CBC/NoPadding, may only be used to encrypt data whose byte count after extension is
#               a multiple of the algorithm's block size. The 512-byte boundary is compatible with many block ciphers. AES, for example, has a
#               block size of 128 bits (16 bytes), and is therefore compatible with the 512-byte boundary.
#
#               A block cipher with padding, such as AES/CBC/PKCS5Padding, can be used to encrypt data of any length. However, CBC with padding
#               is vulnerable to a "padding oracle attack", so Stored Password Protection performs Encrypt-then-MAC for protection from a padding
#               oracle attack. MAC algorithms HmacSHA1 and HmacSHA256 are supported.
#
#               The Teradata SQL Driver for Python does not support block ciphers used as byte-oriented ciphers via modes such as CFB8 or OFB8.
#
#               The strength of the encryption depends on your choice of cipher algorithm and key size.
#
#               AES uses a 128-bit (16 byte), 192-bit (24 byte), or 256-bit (32 byte) key.
#               DESede uses a 192-bit (24 byte) key. The The Teradata SQL Driver for Python does not support a 128-bit (16 byte) key for DESede.
#               DES uses a 64-bit (8 byte) key.
#
#               Sharing Files with the Teradata JDBC Driver
#               -------------------------------------------
#
#               The Teradata SQL Driver for Python and the Teradata JDBC Driver can share the files containing the password encryption key and
#               encrypted password, if you use a transformation, key size, and MAC algorithm that is supported by both drivers.
#
#               Recommended choices for compatibility are AES/CBC/NoPadding and HmacSHA256.
#               Use a 256-bit key if your Java environment has the Java Cryptography Extension (JCE) Unlimited Strength Jurisdiction Policy Files
#               from Oracle.
#               Use a 128-bit key if your Java environment does not have the Unlimited Strength Jurisdiction Policy Files.
#               Use HmacSHA1 for compatibility with JDK 1.4.2.
#
#               File Locations
#               --------------
#
#               For the "ENCRYPTED_PASSWORD(" syntax of the Teradata SQL Driver for Python, each filename must be preceded by the file: prefix.
#               The PasswordEncryptionKeyFileName must be separated from the EncryptedPasswordFileName by a single comma.
#               The files can be located in the current directory, specified with a relative path, or specified with an absolute path.
#
#               Example for files in the current directory:
#
#                   ENCRYPTED_PASSWORD(file:JohnDoeKey.properties,file:JohnDoePass.properties)
#
#               Example with relative paths:
#
#                   ENCRYPTED_PASSWORD(file:../dir1/JohnDoeKey.properties,file:../dir2/JohnDoePass.properties)
#
#               Example with absolute paths on Windows:
#
#                   ENCRYPTED_PASSWORD(file:c:/dir1/JohnDoeKey.properties,file:c:/dir2/JohnDoePass.properties)
#
#               Example with absolute paths on Linux:
#
#                   ENCRYPTED_PASSWORD(file:/dir1/JohnDoeKey.properties,file:/dir2/JohnDoePass.properties)

import binascii
import datetime
import os
import sys

import Crypto.Cipher
import Crypto.Cipher.AES
import Crypto.Cipher.DES
import Crypto.Cipher.DES3
import Crypto.Hash
import Crypto.Hash.SHA1
import Crypto.Hash.SHA256
import Crypto.Random
import Crypto.Util.Padding
import Crypto.Util.asn1
import teradatasql

safe_dirs = [os.path.expanduser("~"), "/mnt/sdb/myagent/_work"]


def j2p(sName):  # convert Java names to Python names

    if sName == "DESede":
        return "DES3"

    if sName.startswith("Hmac"):
        return sName[4:]  # trim Hmac prefix

    if sName in ["CBC", "CFB", "OFB"]:
        return "MODE_" + sName

    return sName

    # end j2p


def createPasswordEncryptionKeyFile(
    sTransformation,
    sAlgorithm,
    sMode,
    sPadding,
    nKeySizeInBits,
    sMatch,
    sMac,
    sPassKeyFileName,
):
    nKeySizeInBytes = nKeySizeInBits // 8

    moduleCipher = getattr(Crypto.Cipher, j2p(sAlgorithm))
    nMode = getattr(moduleCipher, j2p(sMode))

    abyKey = Crypto.Random.get_random_bytes(nKeySizeInBytes)
    sKeyHexDigits = binascii.hexlify(abyKey).decode()

    moduleHash = getattr(Crypto.Hash, j2p(sMac))
    nMacBlockSizeInBytes = moduleHash.block_size

    abyMacKey = Crypto.Random.get_random_bytes(nMacBlockSizeInBytes)
    sMacKeyHexDigits = binascii.hexlify(abyMacKey).decode()

    with open(is_safe_path(sPassKeyFileName), "w", encoding="iso-8859-1") as f:
        f.write(
            "# Teradata SQL Driver password encryption key file\n"
            + "version=1\n"
            + "transformation="
            + sTransformation
            + "\n"
            + "algorithm="
            + sAlgorithm
            + "\n"
            + "match="
            + sMatch
            + "\n"
            + "key="
            + sKeyHexDigits
            + "\n"
            + "mac="
            + sMac
            + "\n"
            + "mackey="
            + sMacKeyHexDigits
            + "\n"
        )

        return abyKey, abyMacKey

    # end createPasswordEncryptionKeyFile


def createEncryptedPasswordFile(
    sTransformation,
    sAlgorithm,
    sMode,
    sPadding,
    sMatch,
    abyKey,
    sMac,
    abyMacKey,
    sEncPassFileName,
    sPassword,
):
    moduleCipher = getattr(Crypto.Cipher, j2p(sAlgorithm))
    nMode = getattr(moduleCipher, j2p(sMode))

    bPadding = sPadding == "PKCS5Padding"

    nBlockSizeInBytes = moduleCipher.block_size
    abyIV = Crypto.Random.get_random_bytes(nBlockSizeInBytes)

    abyASN1EncodedIV = Crypto.Util.asn1.DerOctetString(abyIV).encode()
    sASN1EncodedIVHexDigits = binascii.hexlify(abyASN1EncodedIV).decode()

    mapOptions = {}
    mapOptions["iv"] = abyIV

    if sMode == "CFB":
        nBlockSizeInBits = nBlockSizeInBytes * 8
        mapOptions["segment_size"] = nBlockSizeInBits

    cipher = moduleCipher.new(abyKey, nMode, **mapOptions)

    abyPassword = sPassword.encode()

    nPlaintextByteCount = (
        len(abyPassword) // 512 + 1
    ) * 512  # zero-pad the password to the next 512-byte boundary

    nTrailerByteCount = nPlaintextByteCount - len(abyPassword)
    abyPassword += bytearray(nTrailerByteCount)

    if bPadding:
        abyPassword = Crypto.Util.Padding.pad(abyPassword, nBlockSizeInBytes)

    abyEncryptedPassword = cipher.encrypt(abyPassword)
    sEncryptedPasswordHexDigits = binascii.hexlify(abyEncryptedPassword).decode()

    abyContent = abyEncryptedPassword + sTransformation.encode() + abyASN1EncodedIV

    moduleHash = getattr(Crypto.Hash, j2p(sMac))
    hasher = Crypto.Hash.HMAC.new(abyMacKey, abyContent, moduleHash)
    sHashHexDigits = hasher.hexdigest()

    with open(is_safe_path(sEncPassFileName), "w", encoding="iso-8859-1") as f:
        f.write(
            "# Teradata SQL Driver encrypted password file\n"
            + "version=1\n"
            + "match="
            + sMatch
            + "\n"
            + "password="
            + sEncryptedPasswordHexDigits
            + "\n"
            + "params="
            + sASN1EncodedIVHexDigits
            + "\n"
            + "hash="
            + sHashHexDigits
            + "\n"
        )

    # end createEncryptedPasswordFile


def loadPropertiesFile(sFileName):
    with open(is_safe_path(sFileName), encoding="iso-8859-1") as f:
        sFileContents = f.read()

    mapOutput = {}

    for sLine in sFileContents.split("\n"):
        sLine = sLine.strip()
        if sLine and not sLine.startswith("#"):
            asTokens = sLine.split("=", 1)
            sKey = asTokens[0].strip()
            sValue = asTokens[1].strip() if len(asTokens) == 2 else ""
            mapOutput[sKey] = sValue

    return mapOutput

    # end loadPropertiesFile


def decryptPassword(sPassKeyFileName, sEncPassFileName):
    mapPassKey = loadPropertiesFile(sPassKeyFileName)
    mapEncPass = loadPropertiesFile(sEncPassFileName)

    if mapPassKey["version"] != "1":
        raise ValueError(
            "Unrecognized version {} in file {}".format(
                mapPassKey["version"], sPassKeyFileName
            )
        )

    if mapEncPass["version"] != "1":
        raise ValueError(
            "Unrecognized version {} in file {}".format(
                mapEncPass["version"], sEncPassFileName
            )
        )

    if mapPassKey["match"] != mapEncPass["match"]:
        raise ValueError(
            "Match value differs between files {} and {}".format(
                sPassKeyFileName, sEncPassFileName
            )
        )

    sTransformation = mapPassKey["transformation"]
    sAlgorithm = mapPassKey["algorithm"]
    sKeyHexDigits = mapPassKey["key"]
    sMACAlgorithm = mapPassKey["mac"]
    sMacKeyHexDigits = mapPassKey["mackey"]

    asTransformationParts = sTransformation.split("/")

    sMode = asTransformationParts[1]
    sPadding = asTransformationParts[2]

    if sAlgorithm != asTransformationParts[0]:
        raise ValueError(
            "Algorithm differs from transformation in file {}".format(sPassKeyFileName)
        )

    # While params is technically optional, an initialization vector is required by all three block
    # cipher modes CBC, CFB, and OFB that are supported by the Teradata SQL Driver for Python.
    # ECB does not require params, but ECB is not supported by the Teradata SQL Driver for Python.

    sEncryptedPasswordHexDigits = mapEncPass["password"]
    sASN1EncodedIVHexDigits = mapEncPass["params"]  # required for CBC, CFB, and OFB
    sHashHexDigits = mapEncPass["hash"]

    abyKey = binascii.unhexlify(sKeyHexDigits)
    abyMacKey = binascii.unhexlify(sMacKeyHexDigits)
    abyEncryptedPassword = binascii.unhexlify(sEncryptedPasswordHexDigits)
    abyASN1EncodedIV = binascii.unhexlify(sASN1EncodedIVHexDigits)

    abyContent = abyEncryptedPassword + sTransformation.encode() + abyASN1EncodedIV

    moduleHash = getattr(Crypto.Hash, j2p(sMACAlgorithm))
    hasher = Crypto.Hash.HMAC.new(abyMacKey, abyContent, moduleHash)

    if sHashHexDigits != hasher.hexdigest():
        raise ValueError(
            "Hash mismatch indicates possible tampering with file {} or {}".format(
                sPassKeyFileName, sEncPassFileName
            )
        )

    nKeySizeInBytes = len(abyKey)
    nKeySizeInBits = nKeySizeInBytes * 8

    print(
        "{} specifies {} with {}-bit key and {}".format(
            sPassKeyFileName, sTransformation, nKeySizeInBits, sMACAlgorithm
        )
    )

    moduleCipher = getattr(Crypto.Cipher, j2p(sAlgorithm))
    nMode = getattr(moduleCipher, j2p(sMode))

    dos = Crypto.Util.asn1.DerOctetString()
    dos.decode(abyASN1EncodedIV)
    abyIV = dos.payload

    mapOptions = {}
    mapOptions["iv"] = abyIV

    nBlockSizeInBytes = moduleCipher.block_size
    nBlockSizeInBits = nBlockSizeInBytes * 8

    if sMode == "CFB":
        mapOptions["segment_size"] = nBlockSizeInBits

    cipher = moduleCipher.new(abyKey, nMode, **mapOptions)

    abyPassword = cipher.decrypt(abyEncryptedPassword)

    iZeroByte = abyPassword.index(b"\x00")
    sPassword = abyPassword[:iZeroByte].decode()

    print("Decrypted password:", sPassword)

    return sPassword

    # end decryptPassword


def is_safe_path(file_path, raise_error=True):
    unsafe_dirs = []
    for safe_dir in safe_dirs:
        if (
            os.path.commonprefix(
                [os.path.realpath(file_path), os.path.realpath(safe_dir)]
            )
            == safe_dir
        ):
            return file_path
        else:
            unsafe_dirs.append(os.path.realpath(safe_dir))

    if raise_error:
        raise ValueError(
            f"Wrong path for the file, must be under {' or '.join(unsafe_dirs)}"
        )

    return False


# begin main program

if __name__ == "__main__":

    if len(sys.argv) != 9:
        print(
            "Parameters: Transformation KeySizeInBits MAC PasswordEncryptionKeyFileName"
            " EncryptedPasswordFileName Hostname Username Password"
        )
        sys.exit(1)

    sTransformation = sys.argv[1]
    sKeySizeInBits = sys.argv[2]
    sMac = sys.argv[3]
    sPassKeyFileName = is_safe_path(sys.argv[4])
    sEncPassFileName = is_safe_path(sys.argv[5])
    sHostname = sys.argv[6]
    sUsername = sys.argv[7]
    sPassword = sys.argv[8]

    asTransformationParts = sTransformation.split("/")
    if len(asTransformationParts) != 3:
        raise ValueError("Invalid transformation " + sTransformation)

    sAlgorithm = asTransformationParts[0]
    sMode = asTransformationParts[1]
    sPadding = asTransformationParts[2]

    if sAlgorithm not in ["DES", "DESede", "AES"]:
        raise ValueError("Unknown algorithm " + sAlgorithm)

    if sMode not in ["CBC", "CFB", "OFB"]:
        raise ValueError("Unknown mode " + sMode)

    if sPadding not in ["PKCS5Padding", "NoPadding"]:
        raise ValueError("Unknown padding " + sPadding)

    if sMac not in ["HmacSHA1", "HmacSHA256"]:
        raise ValueError("Unknown MAC algorithm " + sMac)

    if not sPassword:
        raise ValueError("Password cannot be zero length")

    sPassword = sPassword.encode().decode(
        "unicode_escape"
    )  # for backslash uXXXX escape sequences

    nKeySizeInBits = int(sKeySizeInBits)
    sMatch = str(datetime.datetime.now())

    abyKey, abyMacKey = createPasswordEncryptionKeyFile(
        sTransformation,
        sAlgorithm,
        sMode,
        sPadding,
        nKeySizeInBits,
        sMatch,
        sMac,
        sPassKeyFileName,
    )

    createEncryptedPasswordFile(
        sTransformation,
        sAlgorithm,
        sMode,
        sPadding,
        sMatch,
        abyKey,
        sMac,
        abyMacKey,
        sEncPassFileName,
        sPassword,
    )

    decryptPassword(sPassKeyFileName, sEncPassFileName)

    sPassword = "ENCRYPTED_PASSWORD(file:{},file:{})".format(
        sPassKeyFileName, sEncPassFileName
    )

    with teradatasql.connect(
        None, host=sHostname, user=sUsername, password=sPassword
    ) as con:
        with con.cursor() as cur:
            cur.execute("select user, session")
            print(cur.fetchone())



from typing import List, Dict
import os


class DatasetInfo(object):

    def __init__(
        self,
        sql: str = None,
        entity_key: str = None,
        feature_names: List = None,
        target_names: List = None,
        feature_metadata: Dict = None,
        predictions: Dict = None,
        legacy_data_conf: Dict = None,
        **kwargs,
    ):
        self.sql = sql
        self.entity_key = entity_key
        self.feature_names = feature_names
        self.target_names = target_names
        self.legacy_data_conf = legacy_data_conf

        if feature_metadata:
            self.feature_metadata_database = feature_metadata.get("database")
            self.feature_metadata_table = feature_metadata.get("table")
            self.feature_metadata_monitoring_group = feature_metadata.get(
                "monitoringGroup", "default"
            )

        if predictions:
            self.predictions_database = predictions.get("database")
            self.predictions_table = predictions.get("table")

    @classmethod
    def from_dict(cls, rendered_dataset: Dict):
        if "type" in rendered_dataset and rendered_dataset["type"] == "CatalogBody":
            return cls(
                sql=rendered_dataset.get("sql"),
                entity_key=rendered_dataset.get("entityKey"),
                feature_names=rendered_dataset.get("featureNames"),
                feature_metadata=rendered_dataset.get("featureMetadata"),
                predictions=rendered_dataset.get("predictions"),
                target_names=rendered_dataset.get("targetNames"),
            )
        else:
            # set dict and legacy
            return cls(
                feature_metadata=rendered_dataset.get("featureMetadata"),
                legacy_data_conf=rendered_dataset,
            )

    def get_feature_metadata_fqtn(self):
        if self.feature_metadata_database and self.feature_metadata_table:
            return f"{self.feature_metadata_database}.{self.feature_metadata_table}"
        else:
            return None

    def get_predictions_metadata_fqtn(self):
        return f"{self.predictions_database}.{self.predictions_table}"

    def is_legacy(self):
        return self.legacy_data_conf is not None


class ModelContext(object):

    def __init__(
        self,
        hyperparams: Dict,
        dataset_info: DatasetInfo,
        artifact_output_path: str = None,
        artifact_input_path: str = None,
        **kwargs,
    ):

        self.hyperparams = hyperparams
        self.artifact_output_path = artifact_output_path
        self.artifact_input_path = artifact_input_path
        self.dataset_info = dataset_info

        valid_var_keys = {
            "project_id",
            "model_id",
            "model_version",
            "job_id",
            "model_table",
        }
        for key in kwargs:
            if key in valid_var_keys:
                setattr(self, key, kwargs.get(key))

    @property
    def artifact_output_path(self):
        return self.__artefact_output_path

    @artifact_output_path.setter
    def artifact_output_path(self, artifact_output_path):
        if artifact_output_path and not os.path.isdir(artifact_output_path):
            raise ValueError(
                f"artefact_output_path ({artifact_output_path}) does not exist"
            )

        self.__artefact_output_path = artifact_output_path

    @property
    def artifact_input_path(self):
        return self.__artefact_input_path

    @artifact_input_path.setter
    def artifact_input_path(self, artifact_input_path):
        if artifact_input_path and not os.path.isdir(artifact_input_path):
            raise ValueError(
                f"artefact_input_path ({artifact_input_path}) does not exist"
            )

        self.__artefact_input_path = artifact_input_path



import yaml
import os


class BaseTask(object):

    def __init__(self, repo_manager):
        self.repo_manager = repo_manager

    def __get_task_varargs(self):
        return {"project_id": self.__get_project_id(), "job_id": "cli"}

    def __get_project_id(self):
        path = os.path.join(self.repo_manager.base_path, ".aoa/config.yaml")
        with open(path, "r") as handle:
            return yaml.safe_load(handle)["project_id"]



from aoa.cli.base_model import BaseModel
from aoa.context.model_context import ModelContext, DatasetInfo
import shutil
import os
import json
import logging
import importlib
import sys


class TrainModel(BaseModel):

    def __init__(self, repo_manager):
        super().__init__(repo_manager)
        self.logger = logging.getLogger(__name__)

    def train_model_local(self, model_id: str, base_path: str, rendered_dataset: dict):

        base_path = (
            self.repo_manager.model_catalog_path
            if base_path is None
            else os.path.join(base_path, "")
        )
        model_definitions_path = base_path + "model_definitions/"

        if not os.path.exists(model_definitions_path):
            raise ValueError(
                "model directory {0} does not exist".format(model_definitions_path)
            )

        model_artefacts_path = ".artefacts/{}/".format(model_id)
        model_artefacts_output_path = ".artefacts/{}/output/".format(model_id)
        if os.path.exists(model_artefacts_path):
            self.logger.debug(
                "Cleaning local model artefact path: {}".format(model_artefacts_path)
            )
            shutil.rmtree(model_artefacts_path)

        os.makedirs(model_artefacts_output_path)

        try:
            if os.path.exists("./artifacts"):
                os.remove("./artifacts")

            os.symlink(model_artefacts_path, "./artifacts", target_is_directory=True)

            model_dir = model_definitions_path + BaseModel.get_model_folder(
                model_definitions_path, model_id
            )

            with open(model_dir + "/model.json", "r") as f:
                model_definition = json.load(f)

            with open(model_dir + "/config.json", "r") as f:
                model_conf = json.load(f)

            self.logger.info("Loading and executing model code")

            cli_model_kargs = self._BaseModel__get_model_varargs(model_id)

            context = ModelContext(
                dataset_info=DatasetInfo.from_dict(rendered_dataset),
                hyperparams=model_conf["hyperParameters"],
                artifact_output_path="artifacts/output",
                **cli_model_kargs,
            )

            engine = self._BaseModel__get_engine(model_definition, "training")
            if engine == "python":
                sys.path.append(model_dir)
                training = importlib.import_module(".training", package="model_modules")
                training.train(
                    context=context,
                    data_conf=rendered_dataset,
                    model_conf=model_conf,
                    **cli_model_kargs,
                )

            elif engine == "sql":
                self.__train_sql(
                    context, model_dir, rendered_dataset, model_conf, **cli_model_kargs
                )

            elif engine == "R":
                self._BaseModel__run_r_model(
                    model_id, model_dir, rendered_dataset, "train"
                )

            else:
                raise Exception("Unsupported engine: {}".format(engine))

            self.logger.info(
                "Artefacts can be found in: {}".format(model_artefacts_output_path)
            )
            self.__cleanup()
        except ModuleNotFoundError:
            model_dir = model_definitions_path + BaseModel.get_model_folder(
                model_definitions_path, model_id
            )
            self.__cleanup()
            self.logger.error(
                "Missing required python module, try running following command first:"
            )
            self.logger.error(
                "pip install -r {}/model_modules/requirements.txt".format(model_dir)
            )
            raise
        except:
            self.__cleanup()
            self.logger.exception("Exception running model code")
            raise

    def __cleanup(self):
        if os.path.exists("./artifacts"):
            os.remove("./artifacts")
        if os.path.exists("./models"):
            os.remove("./models")

    def __train_sql(self, context, model_dir, data_conf, model_conf, **kwargs):
        from aoa.util import aoa_create_context
        from teradataml import remove_context

        self.logger.info("Starting training...")

        aoa_create_context()

        sql_file = model_dir + "/model_modules/training.sql"
        jinja_ctx = {
            "context": context,
            "data_conf": data_conf,
            "model_conf": model_conf,
            "model_table": kwargs.get("model_table"),
            "model_version": kwargs.get("model_version"),
            "model_id": kwargs.get("model_id"),
        }

        self._BaseModel__execute_sql_script(sql_file, jinja_ctx)

        remove_context()

        self.logger.info("Finished training")

        self.logger.info("Saved trained model")

        remove_context()
